-- Globals --
if not QuickCD then
    QuickCD = {}
end
QuickCD.AlertUI = { frames = {}, framesMoving = {} }

-- Locals --
local originalXDiff = 0
local originalYDiff = 0
local leftDown = false

-- Methods --
function QuickCD.AlertUI.Create(id, moving) -- Creates new alert, saves to QuickCD.AlertUI.frames
    if not QuickCD_Config['alerts'][QuickCD.currRole][id] then
        return false
    end
    local alert = QuickCD.Utils.DeepCopy(QuickCD_Config['alerts'][QuickCD.currRole][id])
    QuickCD.AlertUI.frames[id] = UI.CreateFrame("Text", "", QuickCD.context)
    local frame = QuickCD.AlertUI.frames[id]
    if moving then
        QuickCD.AlertUI.framesMoving[id] = true
    end
    frame:SetVisible(false)
    QuickCD.AlertUI.UpdateFrame(frame, alert, moving)
    frame:SetLayer(99)
    frame:SetMouseMasking('limited')
    if moving then
        QuickCD.AlertUI.framesMoving[id] = true
        QuickCD.AlertUI.SetMoving(frame, moving)
    end
    return true
end

function QuickCD.AlertUI.Remove(id, moving) -- Hides aler tfrom QuickCD.AlertUI.frames. Does not remove as to not cause memory leaks
    if not moving and QuickCD.AlertUI.framesMoving[id] then
        return false
    end
    if moving then
        QuickCD.AlertUI.framesMoving[id] = nil
    end
    if QuickCD.AlertUI.frames[id] ~= nil then
        if moving then
            QuickCD.AlertUI.SetMoving(QuickCD.AlertUI.frames[id], false)
        end
        QuickCD.AlertUI.frames[id]:SetText("")
        QuickCD.AlertUI.frames[id]:SetVisible(false)
        return true
    else
        return false
    end
end

function QuickCD.AlertUI.SetMoving(frame, moving) -- Used when updating and creating frames. Enables/disables all necessary frame movement [via mouse] functionality
    if not moving then
        frame.Event['LeftDown'] = nil
        frame.Event['LeftUp'] = nil
        frame.Event['LeftUpoutside'] = nil
        frame.Event['MouseMove'] = nil
        frame.Event['RightUp'] = nil
    else
        function frame.Event:LeftDown()
            leftDown = true
            local mouse = Inspect.Mouse()
            originalXDiff = mouse.x - frame:GetLeft()
            originalYDiff = mouse.y - frame:GetTop()
            local left, top, right, bottom = frame:GetBounds()
        end
        function frame.Event:LeftUp()
            leftDown = false
        end
        function frame.Event:LeftUpoutside()
            leftDown = false
        end
        function frame.Event:MouseMove(x, y)
            if not leftDown then
                return
            end
            local newX = x - originalXDiff
            local newY = y - originalYDiff
            frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", newX, newY)
        end
        function frame.Event:RightUp()
            QuickCD.AlertUI.SavePosition(id)
        end
        frame:SetText("88")
        frame:SetVisible(true)
    end
end

function QuickCD.AlertUI.Update(id, moving) -- Updates alert to QuickCD.AlertUI.frames if possible; falls back on Create method
    if QuickCD.AlertUI.framesMoving[id] and not moving then -- We don't want to overwrite a moving alert with a non-moving one
        return false
    end
    if QuickCD.AlertUI.frames[id] ~= nil then
        if moving then
            QuickCD.AlertUI.framesMoving[id] = true
        end 
        if not QuickCD_Config['alerts'][QuickCD.currRole][id] then
            return false
        end
        local alert = QuickCD.Utils.DeepCopy(QuickCD_Config['alerts'][QuickCD.currRole][id])
        local frame = QuickCD.AlertUI.frames[id]
        QuickCD.AlertUI.SetMoving(frame, moving)
        QuickCD.AlertUI.UpdateFrame(frame, alert, moving)
        return true
    else
        return QuickCD.AlertUI.Create(id, moving)
    end
end

function QuickCD.AlertUI.UpdateFrame(frame, alert, moving) -- Updates frame settings [position, font color, font size]
    if moving then
        local x = tonumber(QuickCD.CfgUI.xTextField:GetText())
        local y = tonumber(QuickCD.CfgUI.yTextField:GetText())
        local textSize = tonumber(QuickCD.CfgUI.sizeDropdown:GetSelectedItem())
        local colors = QuickCD.Utils.StrSplit(",", QuickCD.CfgUI.colorDropdown:GetSelectedValue())
        local colorR = tonumber(colors[1])
        local colorG = tonumber(colors[2])
        local colorB = tonumber(colors[3])
        if x == nil then
            x = alert['x']
        end
        if y == nil then
            y = alert['y']
        end
        if textSize == nil then
            textSize = alert['textSize']
        end
        if colorR == nil then
            colorR = alert['colorR']
        end
        if colorG == nil then
            colorG = alert['colorG']
        end
        if colorB == nil then
            colorB = alert['colorB']
        end
        frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
        frame:SetFontColor(colorR, colorG, colorB)
        frame:SetFontSize(textSize)
    else
        frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", alert['x'], alert['y'])
        frame:SetFontColor(alert['colorR'], alert['colorG'], alert['colorB'])
        frame:SetFontSize(alert['textSize'])
    end
end

function QuickCD.AlertUI.UpdateCD(id, remaining) -- Updates CD(text) of alert
    if QuickCD.AlertUI.framesMoving[id] then -- Don't update CD when moving
        return false
    end
    if QuickCD.AlertUI.frames[id] ~= nil then
        if remaining == nil then
            remaining = "0"
        end
        local frame = QuickCD.AlertUI.frames[id]
        
        if not QuickCD_Config['on'] then -- Hiding all alerts if QuickCD is turned off
            frame:SetText("")
            frame:SetVisible(false)
            return
        end
        
        if frame:GetVisible() == false then
            frame:SetVisible(true)
        end
        remaining = QuickCD.Utils.ToTimeStr(remaining)
        frame:SetText(remaining)
        return true
    else
        return false
    end
end

function QuickCD.AlertUI.SavePosition(id) -- Saves position of moved alert, and stops it from being moved
    if QuickCD.AlertUI.framesMoving[id] ~= nil then
        local x = string.format("%.0f", QuickCD.AlertUI.frames[id]:GetLeft())
        local y = string.format("%.0f", QuickCD.AlertUI.frames[id]:GetTop())
        QuickCD.CfgUI.xTextField:SetText(x)
        QuickCD.CfgUI.yTextField:SetText(y)
        QuickCD.AlertUI.Remove(id, true)
        originalXDiff = 0
        originalYDiff = 0
        leftDown = false
    end
end

function QuickCD.AlertUI.ClearAllMoving() -- Clears all moving alerts
    for id, value in pairs(QuickCD.AlertUI.framesMoving) do
        QuickCD.AlertUI.Remove(id, true)
    end
    originalXDiff = 0
    originalYDiff = 0
    leftDown = false
end

function QuickCD.AlertUI.Clear() -- Empties QuickCD.AlertUI.frames table
    for id, value in pairs(QuickCD.AlertUI.frames) do
        if QuickCD.AlertUI.framesMoving[id] then
            QuickCD.AlertUI.Remove(id, true)
        else
            QuickCD.AlertUI.Remove(id)
        end
    end
    return true
end