-- Globals --
QuickCD = { currRole = 'r1' }

-- Locals --
local queue = {}
local cache = { abilities = {}, lastCD = {} }
local configDefaults = { alerts = {}, GCD = 1.5, on = true }
local command = nil
local lastUpdate = 0
local updateInterval = .33 -- Seconds
local lastCacheReset = 0
local cacheReset = 30 -- Seconds (Longer = More memory use, less CPU use. Shorter = Less memory use, more CPU use.)
local noRole = false

-- Methods --
function QuickCD.Init(addon) -- Initializing the addon
    if addon ~= "QuickCD" then
        return
    end

    if not QuickCD_Config then
        QuickCD_Config = configDefaults
    end
    
    local roleId = Inspect.TEMPORARY.Role()
    if not roleId then -- Sometimes roleId isn't able to be recieved, ex. after they crash. If this is true, we will default it to 1 and attempt to retrieve it later
        roleId = 1
        noRole = true
    end
    QuickCD.currRole = 'r' .. roleId
    if not QuickCD_Config['alerts'][QuickCD.currRole] then
        QuickCD_Config['alerts'][QuickCD.currRole] = {}
    end
    
    QuickCD.context = UI.CreateContext("QuickCD")
    QuickCD.CfgUI.Init()
    
    table.insert(Event.Ability.Cooldown.Begin, {QuickCD.BeginCD, "QuickCD", "cooldown begin"})
    table.insert(Event.Ability.Cooldown.End, {QuickCD.EndCD, "QuickCD", "cooldown end"})
    table.insert(Event.TEMPORARY.Role, {QuickCD.RoleChanged, "QuickCD", "role changed"}) -- NOTE: There will eventually be an event that replaces this one.
    table.insert(Event.System.Update.Begin, {QuickCD.Update, "QuickCD", "update functions"})
    
    command = LibSlashCmd.Register("quickcd", "QuickCD")
    QuickCD.InitSlashCmds()

    print("QuickCooldowns loaded.")
end

function QuickCD.InitSlashCmds() -- Registering functions with LibSlashCmd
    LibSlashCmd.RegisterFunc(command,
                             "_fallback",
                             function(input)
                                QuickCD.CfgUI.mainFrame:SetVisible(not QuickCD.CfgUI.mainFrame:GetVisible())
                             end)
                             
    LibSlashCmd.RegisterFunc(command,
                             "reset",
                             function(input)
                                if input[0] ~= "confirm" then
                                    print("If you are sure you want to reset your configurations back to default, type: /" .. command .. " reset confirm")
                                else
                                    for key, value in pairs(QuickCD_Config) do
                                        if configDefaults[key] and key ~= 'alerts' then -- Resetting all configs - except alerts - back to default.
                                            QuickCD_Config[key] = configDefaults[key]
                                        end
                                    end
                                    print("Configurations have been reset back to default.")
                                end
                             end)
                             
    LibSlashCmd.RegisterFunc(command,
                             "cacheMode",
                             function(input)
                                local msg = nil
                                if input[0] then
                                    msg = "Cache set to mode: " .. input[0]
                                 end
                                 
                                if input[0] == '1' then
                                    cacheReset = 600
                                    print(msg)
                                elseif input[0] == '2' then
                                    cacheReset = 300
                                    print(msg)
                                elseif input[0] == '3' then
                                    cacheReset = 120
                                    print(msg)
                                elseif input[0] == '4' then
                                    cacheReset = 60
                                    print(msg)
                                elseif input[0] == '5' then
                                    cacheReset = 15
                                    print(msg)
                                elseif input[0] == '6' then
                                    cacheReset = 3
                                    print(msg)
                                elseif input[0] == 'default' then
                                    cacheReset = 30
                                    print(msg)
                                else
                                    print("Incorrect syntax. '/" .. command .. " cacheMode {1/2/3/4/5/6/default}'")
                                end
                                
                                print("WARNING! Do not touch this setting if you do not know what it does. If you already have, type '/" .. command .. " cacheMode default'")
                             end)
end

function QuickCD.Update() -- Runs periodically to do just about everything
    local currTime = Inspect.Time.Frame()
    if currTime - lastUpdate < updateInterval then
        return
    end
    lastUpdate = currTime
    
    if noRole then -- Weren't able to set their actual role initially, let's try to fix it
        QuickCD.RecoverRole()
    end
    
    QuickCD.CleanCache()
    QuickCD.UpdateUI()
        
end

function QuickCD.UpdateUI() -- Update all visual numbers
    local count = #QuickCD_Config['alerts'][QuickCD.currRole]
    for id, alert in pairs(QuickCD_Config['alerts'][QuickCD.currRole]) do
        local abilityName = alert['abilityName']
        if not queue[abilityName] then -- ability not in queue
            QuickCD.StopCD(abilityName, id)
        else
            local ability = queue[abilityName]
            local abilityDetails = QuickCD.GetAbilityDetails(ability, false)
            local timeLeft = QuickCD.GetCDLeft(abilityDetails)
            if timeLeft == nil then
                QuickCD.StopCD(abilityName, id) -- timeLeft is nil, meaning CD has ended
            end
            
            if cache['lastCD'][abilityName] then -- Preventing any skewed CD numbers (ex. bugging out due to GCD)
                local cachedCD = cache['lastCD'][abilityName]['curr']
                if timeLeft >= cachedCD then
                    local currTime = Inspect.Time.Frame()
                    local lastUpdateTime = cache['lastCD'][abilityName]['lastUpdate']
                    local timeDiff = currTime - lastUpdateTime
                    if currTime > lastUpdateTime then
                        local secondsDiff = math.floor(timeDiff)
                        timeLeft = cachedCD - secondsDiff
                        if timeLeft < 0 or (timeLeft <= 0 and cachedCD <= 0) then
                            QuickCD.StopCD(abilityName)
                            return
                        end
                    end
                end
               
                if timeLeft >= cachedCD then -- This will regulate the amount of times the ability is cached; only once per second.
                    return
                end
            end
            
            cache['lastCD'][abilityName] = {curr = timeLeft, lastUpdate = Inspect.Time.Frame()} -- Cache this CD update
            local chk = QuickCD.AlertUI.UpdateCD(id, timeLeft)
            if not chk then -- Alert frame isn't created yet, let's do so.
                QuickCD.AlertUI.Update(id)
                QuickCD.AlertUI.UpdateCD(id, timeLeft)
            end
        end
    end
end

function QuickCD.UpdateQueue(abilityName, ability) -- Updates queue table
    if abilityName == nil or ability == nil then
        return
    end
    queue[abilityName] = ability
end

function QuickCD.BeginCD(cooldowns) -- Cooldown begins
    for ability, timeLeft in pairs(cooldowns) do
        timeLeft = QuickCD.Utils.Round(timeLeft, 1)
        if timeLeft > QuickCD_Config['GCD'] then
            local abilityDetails = QuickCD.GetAbilityDetails(ability)
            local abilityName = abilityDetails.name
            if QuickCD.IsDisabled(abilityName) == false then
                QuickCD.UpdateQueue(abilityName, ability)
            end
        end
    end
end

function QuickCD.EndCD(cooldowns) -- Cooldown ends
    for ability, timeLeft in pairs(cooldowns) do
        local abilityDetails = QuickCD.GetAbilityDetails(ability)
        local abilityName = abilityDetails.name
        QuickCD.StopCD(abilityName)
    end
end

function QuickCD.StopCD(abilityName, id)
    queue[abilityName] = nil    
    cache['lastCD'][abilityName] = nil
    if id ~= nil then
        QuickCD.AlertUI.Remove(id)
    end
end

function QuickCD.IsDisabled(abilityName) -- Is alert set up for ability?
    if abilityName == nil or abilityName == "" then
        return true
    end
    for key, value in pairs(QuickCD_Config['alerts'][QuickCD.currRole]) do
        if value.abilityName == abilityName and value.disabled == false then
            return false
        --else -- Code disabled to prioritize disabled == false over disabled == true
            --return true
        end
    end
    return true
end

function QuickCD.CleanCache() -- Clears ability cache periodically
    local currTime = Inspect.Time.Frame()
    if currTime - lastCacheReset < cacheReset then
        return
    end
    lastCacheReset = currTime
    cache['abilities'] = {}
end

function QuickCD.RoleChanged(roleid) -- Handles player's role changing
    QuickCD.currRole = 'r' .. roleid
    if not QuickCD_Config['alerts'][QuickCD.currRole] then
        QuickCD_Config['alerts'][QuickCD.currRole] = {}
    end
    QuickCD.ClearAll()
end

function QuickCD.RecoverRole() -- If role was not able to be set/found initially (ex. player crashed and came back in), this will help find their actual role ASAP
    local roleId = Inspect.TEMPORARY.Role() -- NOTE: There will eventually be a different method that replaces this one.
    if not roleId then
        return false
    end
    QuickCD.currRole = 'r' .. roleId
    if not QuickCD_Config['alerts'][QuickCD.currRole] then
        QuickCD_Config['alerts'][QuickCD.currRole] = {}
    end
    noRole = false
    return true
end

function QuickCD.GetAbilityDetails(ability, useCache) -- Gets ability details from cache if possible. If not, gets ability details and then caches them. Set useCache to false when getting cooldown remaining
    if useCache == nil then
        useCache = true
    end
    local abilityDetails = nil
    
    if useCache == true and cache['abilities'][ability] then
        abilityDetails = cache['abilities'][ability]
    else
        abilityDetails = Inspect.Ability.Detail(ability)
        cache['abilities'][ability] = abilityDetails
    end
    return abilityDetails
end

function QuickCD.ClearAll() -- Clears everything
    QuickCD.AlertUI.Clear() -- Clear on screen alerts
    queue = {} -- Clear queue
    cache['abilities'] = {} -- Clear cache
    QuickCD.CfgUI.RefreshAlerts() -- Reset/update config alert listings
end

function QuickCD.GetCDLeft(abilityDetails) -- Get's cooldown remaining in integer form, rounded to nearest whole number
    return QuickCD.Utils.Round(abilityDetails.currentCooldownRemaining)
end

-- Event Registration --
table.insert(Event.Addon.Load.End, {QuickCD.Init, "QuickCD", "Initital Setup"})
-- NOTE: Rest of event registration happens in Init method